@@ -1,5 +1,4 @@
 --  C->Haskell Compiler: binding generator
---  vim:ts=8:noexpandtab
 --
 --  Copyright (c) [1999..2003] Manuel M T Chakravarty
 --
@@ -128,7 +127,7 @@ import C2HS.State  (CST, errorsPresent, showErrors, fatal,
                    SwitchBoard(..), Traces(..), putTraceStr, getSwitch)
 import C2HS.C     (AttrC, CObj(..), CTag(..),
                    CDecl(..), CDeclSpec(..), CTypeSpec(..),
-                   CStructUnion(..), CStructTag(..), CEnum(..), CDeclr(..),
+                   CStructUnion(..), CStructTag(..), CEnum(..), CDeclr(..), CAttr(..),
                    CDerivedDeclr(..),CArrSize(..),
                    CExpr(..), CBinaryOp(..), CUnaryOp(..), CConst (..),
                    CInteger(..),cInteger,getCInteger,getCCharAsInt,
@@ -141,11 +140,11 @@ import C2HS.C     (AttrC, CObj(..), CTag(..),
                    checkForAlias, checkForOneAliasName, checkForOneCUName,
                    lookupEnum, lookupStructUnion, lookupDeclOrTag, isPtrDeclr,
                    dropPtrDeclr, isPtrDecl, getDeclOf, isFunDeclr,
-                   refersToNewDef, CDef(..))
+                   refersToNewDef, partitionDeclSpecs, CDef(..))
 
 -- friends
 import C2HS.CHS   (CHSModule(..), CHSFrag(..), CHSHook(..),
-                   CHSParm(..), CHSArg(..), CHSAccess(..), CHSAPath(..),
+                   CHSParm(..), CHSMarsh, CHSArg(..), CHSAccess(..), CHSAPath(..),
                    CHSPtrType(..), showCHSParm, apathToIdent)
 import C2HS.C.Info      (CPrimType(..), alignment, getPlatform)
 import qualified C2HS.C.Info as CInfo
@@ -154,7 +153,6 @@ import C2HS.Gen.Monad    (TransFun, transTabToTransFun, HsObject(..), GB,
                    delayCode, getDelayedCode, ptrMapsTo, queryPtr, objIs,
                    queryClass, queryPointer, mergeMaps, dumpMaps)
 
-
 -- default marshallers
 -- -------------------
 
@@ -166,55 +164,55 @@ import C2HS.Gen.Monad    (TransFun, transTabToTransFun, HsObject(..), GB,
 
 -- | determine the default "in" marshaller for the given Haskell and C types
 --
-lookupDftMarshIn :: String -> [ExtType] -> GB (Maybe (Ident, CHSArg))
+lookupDftMarshIn :: String -> [ExtType] -> GB CHSMarsh
 lookupDftMarshIn "Bool"   [PrimET pt] | isIntegralCPrimType pt =
-  return $ Just (cFromBoolIde, CHSValArg)
+  return $ Just (Left cFromBoolIde, CHSValArg)
 lookupDftMarshIn hsTy     [PrimET pt] | isIntegralHsType hsTy
                                       &&isIntegralCPrimType pt =
-  return $ Just (cIntConvIde, CHSValArg)
+  return $ Just (Left cIntConvIde, CHSValArg)
 lookupDftMarshIn hsTy     [PrimET pt] | isFloatHsType hsTy
                                       &&isFloatCPrimType pt    =
-  return $ Just (cFloatConvIde, CHSValArg)
+  return $ Just (Left cFloatConvIde, CHSValArg)
 lookupDftMarshIn "String" [PtrET (PrimET CCharPT)]             =
-  return $ Just (withCStringIde, CHSIOArg)
+  return $ Just (Left withCStringIde, CHSIOArg)
 lookupDftMarshIn "String" [PtrET (PrimET CCharPT), PrimET pt]
   | isIntegralCPrimType pt                                     =
-  return $ Just (withCStringLenIde, CHSIOArg)
+  return $ Just (Left withCStringLenIde, CHSIOArg)
 lookupDftMarshIn hsTy     [PtrET ty]  | showExtType ty == hsTy =
-  return $ Just (withIde, CHSIOArg)
+  return $ Just (Left withIde, CHSIOArg)
 lookupDftMarshIn hsTy     [PtrET (PrimET pt)]
   | isIntegralHsType hsTy && isIntegralCPrimType pt            =
-  return $ Just (withIntConvIde, CHSIOArg)
+  return $ Just (Left withIntConvIde, CHSIOArg)
 lookupDftMarshIn hsTy     [PtrET (PrimET pt)]
   | isFloatHsType hsTy && isFloatCPrimType pt                  =
-  return $ Just (withFloatConvIde, CHSIOArg)
+  return $ Just (Left withFloatConvIde, CHSIOArg)
 lookupDftMarshIn "Bool"   [PtrET (PrimET pt)]
   | isIntegralCPrimType pt                                     =
-  return $ Just (withFromBoolIde, CHSIOArg)
+  return $ Just (Left withFromBoolIde, CHSIOArg)
 -- FIXME: handle array-list conversion
 lookupDftMarshIn _        _                                    =
   return Nothing
 
 -- | determine the default "out" marshaller for the given Haskell and C types
 --
-lookupDftMarshOut :: String -> [ExtType] -> GB (Maybe (Ident, CHSArg))
+lookupDftMarshOut :: String -> [ExtType] -> GB CHSMarsh
 lookupDftMarshOut "()"     _                                    =
-  return $ Just (voidIde, CHSVoidArg)
+  return $ Just (Left voidIde, CHSVoidArg)
 lookupDftMarshOut "Bool"   [PrimET pt] | isIntegralCPrimType pt =
-  return $ Just (cToBoolIde, CHSValArg)
+  return $ Just (Left cToBoolIde, CHSValArg)
 lookupDftMarshOut hsTy     [PrimET pt] | isIntegralHsType hsTy
                                        &&isIntegralCPrimType pt =
-  return $ Just (cIntConvIde, CHSValArg)
+  return $ Just (Left cIntConvIde, CHSValArg)
 lookupDftMarshOut hsTy     [PrimET pt] | isFloatHsType hsTy
                                        &&isFloatCPrimType pt    =
-  return $ Just (cFloatConvIde, CHSValArg)
+  return $ Just (Left cFloatConvIde, CHSValArg)
 lookupDftMarshOut "String" [PtrET (PrimET CCharPT)]             =
-  return $ Just (peekCStringIde, CHSIOArg)
+  return $ Just (Left peekCStringIde, CHSIOArg)
 lookupDftMarshOut "String" [PtrET (PrimET CCharPT), PrimET pt]
   | isIntegralCPrimType pt                                      =
-  return $ Just (peekCStringLenIde, CHSIOArg)
+  return $ Just (Left peekCStringLenIde, CHSIOArg)
 lookupDftMarshOut hsTy     [PtrET ty]  | showExtType ty == hsTy =
-  return $ Just (peekIde, CHSIOArg)
+  return $ Just (Left peekIde, CHSIOArg)
 -- FIXME: add combination, such as "peek" plus "cIntConv" etc
 -- FIXME: handle array-list conversion
 lookupDftMarshOut _        _                                    =
@@ -272,20 +270,24 @@ voidIde, cFromBoolIde, cToBoolIde, cIntConvIde, cFloatConvIde,
   withFloatConvIde, withFromBoolIde, peekIde, peekCStringIde,
   peekCStringLenIde :: Ident
 voidIde           = internalIdent "void"         -- never appears in the output
-cFromBoolIde      = internalIdent "cFromBool"
-cToBoolIde        = internalIdent "cToBool"
-cIntConvIde       = internalIdent "cIntConv"
-cFloatConvIde     = internalIdent "cFloatConv"
+cFromBoolIde      = internalIdent "fromBool"
+cToBoolIde        = internalIdent "toBool"
+cIntConvIde       = internalIdent "fromIntegral"
+cFloatConvIde     = internalIdent "realToFrac"
 withIde           = internalIdent "with"
 withCStringIde    = internalIdent "withCString"
-withCStringLenIde = internalIdent "withCStringLenIntConv"
-withIntConvIde    = internalIdent "withIntConv"
-withFloatConvIde  = internalIdent "withFloatConv"
-withFromBoolIde   = internalIdent "withFromBoolConv"
+withCStringLenIde = internalIdent "withCStringLenIntConv" --TODO: kill off
+withIntConvIde    = internalIdent "withIntConv"           --TODO: kill off
+withFloatConvIde  = internalIdent "withFloatConv"         --TODO: kill off
+withFromBoolIde   = internalIdent "withFromBoolConv"      --TODO: kill off
 peekIde           = internalIdent "peek"
 peekCStringIde    = internalIdent "peekCString"
-peekCStringLenIde = internalIdent "peekCStringLenIntConv"
+peekCStringLenIde = internalIdent "peekCStringLenIntConv" --TODO: kill off
 
+--TODO: c2hs should not generate these references to externally defined
+-- non-standard utility functions. It's annoying and they are all trivial.
+-- The solutionis to generate expressions inline, rather than requiring all
+-- marshalers be single identifiers.
 
 -- expansion of binding hooks
 -- --------------------------
@@ -404,6 +406,19 @@ expandHook (CHSType ide pos) =
     traceInfoDump decl ty = traceGenBind $
       "Declaration\n" ++ show decl ++ "\ntranslates to\n"
       ++ showExtType ty ++ "\n"
+expandHook (CHSAlignof ide _) =
+  do
+    traceInfoAlignof
+    decl <- findAndChaseDecl ide False True     -- no indirection, but shadows
+    (_, align) <- sizeAlignOf decl
+    traceInfoDump (render $ pretty decl) align
+    return $ show align
+  where
+    traceInfoAlignof         = traceGenBind "** alignment hook:\n"
+    traceInfoDump decl align = traceGenBind $
+      "Alignment of declaration\n" ++ show decl ++ "\nis "
+      ++ show align ++ "\n"
+
 expandHook (CHSSizeof ide _) =
   do
     traceInfoSizeof
@@ -769,7 +784,8 @@ callImport hook isPure isUns ideLexeme hsLexeme cdecl pos =
     extType <- extractFunType pos cdecl isPure
     header  <- getSwitch headerSB
     when (isVariadic extType) (variadicErr pos (posOf cdecl))
-    delayCode hook (foreignImport header ideLexeme hsLexeme isUns extType)
+    delayCode hook (foreignImport (extractCallingConvention cdecl)
+                    header ideLexeme hsLexeme isUns extType)
     traceFunType extType
   where
     traceFunType et = traceGenBind $
@@ -791,9 +807,10 @@ callImportDyn hook _isPure isUns ideLexeme hsLexeme ty pos =
 
 -- | Haskell code for the foreign import declaration needed by a call hook
 --
-foreignImport :: String -> String -> String -> Bool -> ExtType -> String
-foreignImport header ident hsIdent isUnsafe ty  =
-  "foreign import ccall " ++ safety ++ " " ++ show entity ++
+foreignImport :: CallingConvention -> String -> String -> String -> Bool -> ExtType -> String
+foreignImport cconv header ident hsIdent isUnsafe ty  =
+  "foreign import " ++ showCallingConvention cconv ++ " " ++ safety
+  ++ " " ++ show entity ++
   "\n  " ++ hsIdent ++ " :: " ++ showExtType ty ++ "\n"
   where
     safety = if isUnsafe then "unsafe" else "safe"
@@ -856,15 +873,19 @@ funDef isPure hsLexeme fiLexeme extTy octxt parms parm marsh2 pos =
                                    join (take 1 callArgs) ++
                                    " >>= \\b1' ->\n"
       marshRes  = case parm' of
-                    CHSParm _ _ _twoCVal (Just (_    , CHSVoidArg)) _ -> ""
-                    CHSParm _ _ _twoCVal (Just (omIde, CHSIOVoidArg)) _ ->
-                      "  " ++ identToString omIde ++ " res >> \n"
-                    CHSParm _ _ _twoCVal (Just (omIde, CHSIOArg  )) _ ->
-                      "  " ++ identToString omIde ++ " res >>= \\res' ->\n"
-                    CHSParm _ _ _twoCVal (Just (omIde, CHSValArg )) _ ->
-                      "  let {res' = " ++ identToString omIde ++ " res} in\n"
+                    CHSParm _ _ _twoCVal (Just (_     , CHSVoidArg  )) _ -> ""
+                    CHSParm _ _ _twoCVal (Just (omBody, CHSIOVoidArg)) _ ->
+                      "  " ++ marshBody omBody ++ " res >> \n"
+                    CHSParm _ _ _twoCVal (Just (omBody, CHSIOArg     )) _ ->
+                      "  " ++ marshBody omBody ++ " res >>= \\res' ->\n"
+                    CHSParm _ _ _twoCVal (Just (omBody, CHSValArg    )) _ ->
+                      "  let {res' = " ++ marshBody omBody ++ " res} in\n"
                     CHSParm _ _ _       Nothing                    _ ->
                       interr "GenBind.funDef: marshRes: no default?"
+
+      marshBody (Left ide) = identToString ide
+      marshBody (Right str) = str
+
       retArgs'  = case parm' of
                     CHSParm _ _ _ (Just (_, CHSVoidArg))   _ ->        retArgs
                     CHSParm _ _ _ (Just (_, CHSIOVoidArg)) _ ->        retArgs
@@ -912,11 +933,11 @@ funDef isPure hsLexeme fiLexeme extTy octxt parms parm marsh2 pos =
     -- for an argument marshaller, generate all "in" and "out" marshalling
     -- code fragments
     --
-    marshArg i (CHSParm (Just (imIde, imArgKind)) _ twoCVal
-                        (Just (omIde, omArgKind)) _        ) =
+    marshArg i (CHSParm (Just (imBody, imArgKind)) _ twoCVal
+                        (Just (omBody, omArgKind)) _        ) =
       let
         a        = "a" ++ show (i :: Int)
-        imStr    = identToString imIde
+        imStr    = marshBody imBody
         imApp    = imStr ++ " " ++ a
         funArg   = if imArgKind == CHSVoidArg then "" else a
         inBndr   = if twoCVal
@@ -930,7 +951,7 @@ funDef isPure hsLexeme fiLexeme extTy octxt parms parm marsh2 pos =
         callArgs = if twoCVal
                      then [a ++ "'1 ", a ++ "'2"]
                      else [a ++ "'"]
-        omApp    = identToString omIde ++ join callArgs
+        omApp    = marshBody omBody ++ " " ++ join callArgs
         outBndr  = a ++ "''"
         marshOut = case omArgKind of
                      CHSVoidArg   -> ""
@@ -939,6 +960,9 @@ funDef isPure hsLexeme fiLexeme extTy octxt parms parm marsh2 pos =
                      CHSValArg    -> "let {" ++ outBndr ++ " = " ++
                                    omApp ++ "} in "
         retArg   = if omArgKind == CHSVoidArg || omArgKind == CHSIOVoidArg then "" else outBndr
+
+        marshBody (Left ide) = identToString ide
+        marshBody (Right str) = str
       in
       (funArg, marshIn, callArgs, marshOut, retArg)
     marshArg _ _ = interr "GenBind.funDef: Missing default?"
@@ -1027,7 +1051,7 @@ addDftMarshaller pos parms parm extTy = do
     --
     addDftVoid marsh@(Just (_, kind)) = return (marsh, kind == CHSIOArg)
     addDftVoid        Nothing         = do
-      return (Just (internalIdent "void", CHSVoidArg), False)
+      return (Just (Left (internalIdent "void"), CHSVoidArg), False)
 
 -- | compute from an access path, the declarator finally accessed and the index
 -- path required for the access
@@ -1703,6 +1727,39 @@ specType cpos specs'' osize =
               int    = CIntType    undefined
               signed = CSignedType undefined
 
+-- handle calling convention
+-- -------------------------
+
+data CallingConvention = StdCallConv
+                       | CCallConv
+                       deriving (Eq)
+
+-- | determine the calling convention for the provided decl
+extractCallingConvention :: CDecl -> CallingConvention
+extractCallingConvention cdecl
+  | hasStdCallAttr cdecl = StdCallConv
+  | otherwise            = CCallConv
+  where
+    isStdCallAttr (CAttr x _ _) = identToString x == "stdcall"
+                               || identToString x == "__stdcall__"
+
+    hasStdCallAttr = any isStdCallAttr . funAttrs
+
+    funAttrs (CDecl specs declrs _) =
+      let (_,attrs',_,_,_) = partitionDeclSpecs specs
+       in attrs' ++ funEndAttrs declrs
+
+    -- attrs after the function name, e.g. void foo() __attribute__((...));
+    funEndAttrs [(Just ((CDeclr _ (CFunDeclr _ _ _ : _) _ attrs _)), _, _)] = attrs
+    funEndAttrs _                                                           = []
+
+
+-- | generate the necessary parameter for "foreign import" for the
+-- provided calling convention
+showCallingConvention :: CallingConvention -> String
+showCallingConvention StdCallConv = "stdcall"
+showCallingConvention CCallConv   = "ccall"
+
 
 -- offset and size computations
 -- ----------------------------
@@ -1778,6 +1835,7 @@ sizeAlignOfStruct decls CStructTag  =
         align'        = if align > 0 then align else bitfieldAlignment
         alignOfStruct = preAlign `max` align'
     return (sizeOfStruct, alignOfStruct)
+
 sizeAlignOfStruct decls CUnionTag   =
   do
     PlatformSpec {bitfieldAlignmentPS = bitfieldAlignment} <- getPlatform
